"""Utility functions"""
import os

__all__ = ('sub_env',)


def sub_env(string: str):
    """Substitutes environment variables of the form {VARIABLE_NAME} in the
    string"""
    return string.format(**os.environ)



"""
Abstract base class for checks
"""
import typing as t

__all__ = ('CheckBase',)


class CheckBase:
    """Check base class and grouper"""

    # Name of the check
    name: str

    # Value for the check to check
    value: t.Any

    # Description of the check
    desc: str = ''

    # The check (and sub-checks) are enabled
    enabled: bool = True

    # The message to print during the check. {STATUS} is substituted
    msg: str = ''

    # A list of sub checks
    sub_checks: list

    def __init__(self,
                 name: str,
                 value: t.Any,
                 desc: str = '',
                 sub_checks: t.Optional[list['CheckBase', ...]] = None):
        # Make sure the sub_checks are checks
        if sub_checks is not None:
            msg = "All sub-checkers must be instances of CheckBase"
            assert all(isinstance(check, CheckBase)
                       for check in sub_checks), msg
        else:
            sub_checks = []

        # Set attributes
        self.name = name
        self.value = value
        self.desc = desc
        self.sub_checks = list(sub_checks)

    def check(self) -> bool:
        """Performs the checks and sub-checks.

        Returns
        -------
        passed
            True if the check and all sub-checks passed
            False if the check or any sub-checked failed
        """
        # Run all subchecks
        return all(sub.check() for sub in self.sub_checks)

    @classmethod
    def types_dict(cls) -> t.Dict[str, type]:
        """Return all the types and subtypes of Checks in a dict."""
        types = [cls] + cls.__subclasses__()
        return {c.__name__:c for c in types}



"""Constructs for dict types"""
import typing as t

from ..checks import CheckBase

__all__ = ('construct_from_dict', )


def construct_from_dict(name: str,
                        d: dict,
                        level: int = 1,
                        max_level: int = 10) -> t.Union[None, CheckBase]:
    """Construct a Check tree from a dict.

    Notes
    -----
    This function is meant to work recursively."""
    # Check that the maximum recursion level hasn't been reached
    if level >= max_level:
        msg = f"Parsing level {level} exceed the maximum level of {max_level}"
        raise NotImplementedError(msg)

    # Get a listing of the available Check types
    check_types = CheckBase.types_dict()

    # See if the dict makes a call to a Check type
    matching_keys = [k for k in d.keys() if k in check_types]
    if len(matching_keys) > 1:
        msg = f"More than 1 check type specified: {matching_keys}"
        raise NotImplementedError(msg)

    # Parse the check if a single check was given
    if len(matching_keys) == 1:
        # Get the check class
        check_type = matching_keys[0]
        cls = check_types[check_type]

        # Get the value for the check
        value = d[check_type]

        # Get the other kwargs
        kwargs = {k: v for k, v in d.items() if k != check_type}

        # Create and return the check_type
        return cls(name, value, **kwargs)

    # Otherwise, try parsing the sub_checks
    items = d.items()
    found_checks = []  # Values parsed into CheckBase objects
    other_d = dict()   # All other values
    for key, value in items:
        if not isinstance(value, dict):
            other_d[key] = value
            continue

        return_value = construct_from_dict(name=key, d=value, level=level + 1,
                                           max_level=max_level)

        # Replace the value withe CheckBase instance, if it was parsed correctly
        # Otherwise, just place it in the parsed_dict.
        if isinstance(return_value, CheckBase):
            found_checks.append(return_value)
        else:
            other_d[key] = value

    # No sub-checks found; nothing else to do
    if len(found_checks) == 0:
        return None

    # Create a check grouping, first, by parsing the other arguments
    return CheckBase(name=name, sub_checks=found_checks, **other_d)




"""Messages for the command-line interface"""
import typing as t
import enum
import sys
import os

__all__ = ('p_header', 'p_pass', 'p_fail', 'p_bold', 'p_warn')


class Colors(enum.StrEnum):
    """ASCII terminal colors"""
    RED = "\033[1;31m"
    BLUE = "\033[1;34m"
    CYAN = "\033[1;36m"
    GREEN = "\033[0;32m"
    YELLOW = "\033[1;33m"
    MAGENTA = "\033[1;35m"
    RESET = "\033[0;0m"
    BOLD = "\033[;1m"
    REVERSE = "\033[;7m"


def p_header(msg: str, end: str = '\n'):
    """Print a heading"""
    msg = msg.strip() + ' '
    term_width = os.get_terminal_size()[0]

    sys.stdout.write(f'{Colors.MAGENTA}{msg:=<{term_width - 1}}{Colors.RESET}'
                     f'{end}')


def p_pass(msg: str, end: str = '\n'):
    """Print a message for a passed test"""
    sys.stdout.write(f'  {Colors.GREEN}✔ {msg.strip()}{Colors.RESET}{end}')


def p_fail(msg: str, end: str = '\n'):
    """Print a message for a failed test"""
    sys.stderr.write(f'  {Colors.RED}✖ {msg.strip()}{Colors.RESET}{end}')


def p_bold(msg: str, end: str = '\n'):
    """Print a message in bold"""
    sys.stdout.write(f'  {Colors.BOLD}  {msg.strip()}{Colors.RESET}{end}')


def p_warn(msg: str, end: str = '\n'):
    """Print a message for a warning"""
    sys.stderr.write(f'  {Colors.YELLOW}! {msg.strip()}{Colors.RESET}{end}')


"""
Test the environment variable check class
"""
import os

from pytest import MonkeyPatch

from envalidate.checks import CheckEnv


def test_check_env_missing(variable_name='missingVariable'):
    """Test CheckEnv for a missing environment variable"""
    assert variable_name not in os.environ

    check = CheckEnv(name='missing', value=variable_name)
    assert not check.check()


def test_check_env_present(variable_name='PATH'):
    """Test CheckEnv for a present environment variable"""
    assert variable_name in os.environ

    check = CheckEnv(name='present', value=variable_name)
    assert check.check()


def test_check_env_name_substitution(variable_name='VARIABLE_NAME',
                                     variable_value='PATH'):
    """Test CheckEnv with substituting variable names with environment
    variables"""
    # Make sure the variable name isn't in the environment already
    assert variable_name not in os.environ
    assert '{' + variable_name + '}' not in os.environ

    # Set the VariableName envinronment variable to be the name of another
    # environment variable that exists
    with MonkeyPatch.context() as mp:
        mp.setenv(variable_name, variable_value)

        # Setup the check to use the variable name with environment variable
        # substitution--i.e. wrap it in curly braces
        check = CheckEnv(name='substitution',
                         value='{' + variable_name + '}')

        # Check that the subsitution is correct
        assert check.variable_name == 'PATH'
        assert check.value == '{' + variable_name + '}'  # hidden var

        assert check.check()


def test_check_env_regex(variable_name='VARIABLE_NAME',
                         variable_value='dev',
                         regex='dev|prod|local'):
    # Make sure the variable name isn't in the environment already
    assert variable_name not in os.environ

    # Place the variable in the environment and test the regex
    with MonkeyPatch.context() as mp:
        mp.setenv(variable_name, variable_value)

        # The regex should match
        check1 = CheckEnv(name='regex1', value=variable_name,
                          regex=regex)
        assert check1.check()

        # Change the variable value to something else in which the regex will
        # fail
        mp.setenv(variable_name, '!' + variable_value + '!')

        # The regex should not match
        check2 = CheckEnv(name='regex2', value=variable_name,
                          regex=regex)
        assert not check2.check()



"""
Checks environment variables
"""
import typing as t
import os
import re

from .base import CheckBase
from .utils import sub_env
from ..cli import p_pass, p_fail


class CheckEnv(CheckBase):
    """Check the current environment variables.

    Notes
    -----
    This function can substitute variable names and values with other
    environment variables, which is common with environment variables.

    e.g. With the following environment variables:
    ENV=DEV
    DEPLOYMENT_DEV=passed

    The user could use the variable name 'DEPLOYMENT_{ENV}' to match
    'DEPLOYMENT_DEV'.
    """

    value: str

    # (Optional) regex to match the environment variable value
    regex: t.Optional[t.Tuple[str, ...]] = None

    # If True (default), environment variables in variable_name or
    # variable_value are substituted with the values of other environment
    # variables.
    env_substitute: bool = True

    # The message for checking environment variables
    msg: str = "Check environment variable '{variable_name}'...{status}."

    def __init__(self,
                 *args,
                 regex: t.Optional[str] = None,
                 env_substitute: t.Optional[bool] = None,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.regex = regex

        if isinstance(env_substitute, bool):
            self.env_substitute = env_substitute

    @property
    def variable_name(self) -> str:
        """The name of the environment variable with environment variable
        substitution"""
        return (sub_env(self.value) if self.env_substitute else
                self.value)

    @variable_name.setter
    def variable_name(self, value: str) -> None:
        self.value = value

    @property
    def variable_value(self) -> t.Union[None, str]:
        value = os.environ.get(self.variable_name, None)
        if value is not None and self.env_substitute:
            return sub_env(value)
        else:
            return value

    def check(self) -> bool:
        """Check the environment variable value."""
        # Substitute environment variables, if needed
        variable_name = self.variable_name
        variable_value = self.variable_value

        # Make sure the environment variable exists.
        if variable_value is None:
            msg = self.msg.format(variable_name=variable_name,
                                  status='missing')
            p_fail(msg)
            return False

        # Check that the variable has a non-zero value
        if variable_value == '':
            msg = self.msg.format(variable_name=variable_name,
                                  status='empty string')
            p_fail(msg)
            return False

        # Check the regex, if specified
        if (isinstance(self.regex, str) and
            re.match(self.regex, variable_value) is None):
            status = ("value does not match regex "
                      "'{regex}'".format(regex=self.regex))
            msg = self.msg.format(variable_name=variable_name, status=status)
            p_fail(msg)
            return False

        # All checks passed!
        msg = self.msg.format(variable_name=variable_name, status='passed')
        p_pass(msg)
        return True
